"""
Chat app wrapper with multiple local histories.
Run: python -m scripts.examples.chat_app_with_histories.main
"""

from dataclasses import dataclass
from datetime import datetime, timezone
import inspect
import json
import os
from pathlib import Path
from uuid import uuid4

from fasthtml.common import *
from fastsql import Database
from monsterui.all import (
    Button,
    ButtonT,
    Container,
    ContainerT,
    FastHTML as MUFastHTML,
    TextPresets,
    Theme,
    UkIcon,
)

from pylogue.core import (
    EchoResponder,
    IMPORT_PREFIX,
    _register_google_auth_routes,
    _session_cookie_name,
    google_oauth_config_from_env,
    get_core_headers,
    register_core_static,
    register_ws_routes,
    render_cards,
    render_input,
)
from starlette.requests import Request
from starlette.responses import FileResponse
from starlette.responses import JSONResponse
from starlette.responses import RedirectResponse

PROJECT_ROOT = Path(__file__).resolve().parents[2]
CHAT_APP_DIR = PROJECT_ROOT / "scripts" / "examples" / "chat_app_with_histories"
STATIC_DIR = Path(__file__).resolve().parent / "static"
DB_PATH = CHAT_APP_DIR / "chat_app.db"


@dataclass
class Chat:
    id: str
    title: str
    created_at: str
    updated_at: str
    payload: str = ""


def _utc_iso() -> str:
    return datetime.now(timezone.utc).replace(microsecond=0).isoformat()


def app_factory(
    responder=None,
    responder_factory=None,
    db_path: Path | str | None = None,
    sidebar_title: str = "Pylogue",
    hero_title: str = "Fast HTML + Pylogue Core",
    hero_subtitle: str = (
        "One UI wraps multiple Pylogue chat sessions. Pick a chat on the left, "
        "start a new one, or return to previous conversations instantly."
    ),
) -> MUFastHTML:
    if db_path is None:
        caller_frame = inspect.stack()[1]
        caller_file = Path(caller_frame.filename).resolve()
        caller_name = caller_file.stem
        resolved_db_path = caller_file.parent / f"{caller_name}.db"
    else:
        resolved_db_path = Path(db_path)
    local_db = Database(f"sqlite:///{resolved_db_path}")
    if responder_factory is None:
        responder = responder or EchoResponder()
    headers = list(get_core_headers(include_markdown=True))
    headers.extend(
        [
            Link(
                rel="stylesheet",
                href="https://fonts.googleapis.com/css2?family=Martel+Sans:wght@300;400;500;600;700&display=swap",
            ),
            Link(rel="stylesheet", href="/static/chat_app.css"),
            Script(src="/static/chat_app.js", type="module"),
        ]
    )

    oauth_cfg = google_oauth_config_from_env()
    auth_required = bool(oauth_cfg and oauth_cfg.auth_required)
    session_secret = (
        oauth_cfg.session_secret
        if oauth_cfg and oauth_cfg.session_secret
        else os.getenv("PYLOGUE_SESSION_SECRET")
    )
    app_kwargs = {"exts": "ws", "hdrs": tuple(headers), "pico": False}
    app_kwargs["session_cookie"] = _session_cookie_name()
    if session_secret:
        app_kwargs["secret_key"] = session_secret
    app = MUFastHTML(**app_kwargs)
    register_core_static(app)
    auth_paths = _register_google_auth_routes(app, oauth_cfg) if oauth_cfg else None

    def _is_authorized(request: Request) -> bool:
        if not auth_required:
            return True
        auth = request.session.get("auth")
        return isinstance(auth, dict)

    @app.route("/static/chat_app.css")
    def _chat_app_css():
        return FileResponse(STATIC_DIR / "chat_app.css")

    @app.route("/static/chat_app.js")
    def _chat_app_js():
        return FileResponse(STATIC_DIR / "chat_app.js")

    @app.route("/api/chats", methods=["GET"])
    def list_chats(request: Request):
        if not _is_authorized(request):
            return JSONResponse({"error": "Unauthorized"}, status_code=401)
        items = list(local_db.create(Chat, pk="id")())
        items.sort(key=lambda c: c.updated_at or c.created_at, reverse=True)
        return JSONResponse(
            [
                {
                    "id": c.id,
                    "title": c.title,
                    "created_at": c.created_at,
                    "updated_at": c.updated_at,
                }
                for c in items
            ]
        )

    @app.route("/api/chats", methods=["POST"])
    async def create_chat(request: Request):
        if not _is_authorized(request):
            return JSONResponse({"error": "Unauthorized"}, status_code=401)
        chats = local_db.create(Chat, pk="id")
        data = await request.json()
        chat_id = data.get("id") or str(uuid4())
        title = data.get("title") or "New chat"
        now = _utc_iso()
        payload = data.get("payload")
        payload_str = json.dumps(payload) if payload is not None else ""
        chat = Chat(chat_id, title, now, now, payload_str)
        try:
            _ = chats[chat_id]
            chats.update(chat)
        except Exception:
            chats.insert(chat)
        return JSONResponse(
            {
                "id": chat.id,
                "title": chat.title,
                "created_at": chat.created_at,
                "updated_at": chat.updated_at,
            }
        )

    @app.route("/api/chats/{chat_id}", methods=["GET"])
    def get_chat(request: Request, chat_id: str):
        if not _is_authorized(request):
            return JSONResponse({"error": "Unauthorized"}, status_code=401)
        chats = local_db.create(Chat, pk="id")
        try:
            chat = chats[chat_id]
        except Exception:
            return JSONResponse({"cards": []})
        payload = chat.payload or ""
        if not payload:
            return JSONResponse({"cards": []})
        try:
            data = json.loads(payload)
        except json.JSONDecodeError:
            data = {"cards": []}
        return JSONResponse(data)

    @app.route("/api/chats/{chat_id}", methods=["POST"])
    async def save_chat(chat_id: str, request: Request):
        if not _is_authorized(request):
            return JSONResponse({"error": "Unauthorized"}, status_code=401)
        chats = local_db.create(Chat, pk="id")
        data = await request.json()
        payload = data.get("payload") or {"cards": []}
        title = data.get("title") or "New chat"
        now = _utc_iso()
        try:
            existing = chats[chat_id]
            created_at = existing.created_at
        except Exception:
            created_at = data.get("created_at") or now
        chat = Chat(chat_id, title, created_at, now, json.dumps(payload))
        try:
            _ = chats[chat_id]
            chats.update(chat)
        except Exception:
            chats.insert(chat)
        return JSONResponse(
            {
                "id": chat.id,
                "title": chat.title,
                "created_at": chat.created_at,
                "updated_at": chat.updated_at,
            }
        )

    @app.route("/api/chats/{chat_id}", methods=["DELETE"])
    def delete_chat(request: Request, chat_id: str):
        if not _is_authorized(request):
            return JSONResponse({"error": "Unauthorized"}, status_code=401)
        chats = local_db.create(Chat, pk="id")
        try:
            chats.delete(chat_id)
        except Exception:
            pass
        return JSONResponse({"deleted": True})

    sessions: dict[int, dict] = {}
    register_ws_routes(
        app,
        responder=responder,
        responder_factory=responder_factory,
        sessions=sessions,
        auth_required=auth_required,
    )

    def _sidebar(request: Request):
        show_logout = auth_required and _is_authorized(request)
        logout_href = auth_paths["logout_path"] if auth_paths else "/logout"
        return Div(
            Div(
                H1(sidebar_title, cls="text-xl font-semibold"),
                Div(
                    Button(
                        "Close",
                        type="button",
                        id="close-sidebar-btn",
                        cls=(ButtonT.secondary, "text-xs mobile-only-inline"),
                    ),
                    A(
                        UkIcon("sign-out"),
                        Span("Logout"),
                        href=logout_href,
                        cls=(ButtonT.secondary, "text-xs"),
                    ) if show_logout else None,
                    cls="flex items-center gap-2",
                ),
                cls="sidebar-header",
            ),
            Button(
                # UkIcon("plus"),
                Span("New chat"),
                cls=(ButtonT.secondary, "w-full justify-center"),
                type="button",
                id="new-chat-btn",
            ),
            Div(id="chat-list", cls="chat-list"),
            cls="sidebar",
        )

    def _hero():
        return Div(
            Div(
                H3(hero_title, cls="hero-title"),
                P(hero_subtitle, cls="hero-sub"),
                cls="space-y-2",
            ),
            Div(
                Button(
                    "Chats",
                    type="button",
                    id="sidebar-toggle-btn",
                    cls=(ButtonT.secondary, "mobile-only-inline"),
                ),
                Button(
                    UkIcon("download"),
                    cls="uk-button uk-button-text copy-chat-btn",
                    type="button",
                    aria_label="Download conversation JSON",
                    title="Download conversation JSON",
                ),
                Button(
                    UkIcon("upload"),
                    cls="uk-button uk-button-text copy-chat-btn upload-chat-btn",
                    type="button",
                    aria_label="Upload conversation JSON",
                    title="Upload conversation JSON",
                ),
                Input(
                    type="file",
                    id="chat-upload",
                    accept="application/json",
                    cls="sr-only",
                ),
                cls="flex gap-2 justify-end",
            ),
            cls="hero",
        )

    def _chat_content():
        return Div(
            render_cards([]),
            cls="chat-content-wrapper",
        )

    def _chat_input():
        return Div(
            Form(
                render_input(),
                Div(
                    Button("Send", cls=ButtonT.primary, type="submit", id="chat-send-btn"),
                    cls="flex items-center",
                ),
                id="form",
                hx_ext="ws",
                ws_connect="/ws",
                ws_send=True,
                hx_target="#cards",
                hx_swap="outerHTML",
                cls="flex flex-col sm:flex-row gap-3 items-stretch",
            ),
            cls="chat-input-area",
        )

    def _main_panel():
        return Div(
            _hero(),
            _chat_content(),
            _chat_input(),
            cls="main-panel",
        )

    def _delete_chat_modal():
        return Div(
            Div(
                H3("Delete chat?", cls="delete-chat-modal-title"),
                P(
                    "This will permanently delete ",
                    Strong("", id="delete-chat-name"),
                    ".",
                    cls="delete-chat-modal-text",
                ),
                Div(
                    Button(
                        "Cancel",
                        type="button",
                        id="delete-chat-cancel-btn",
                        cls=(ButtonT.secondary, "delete-chat-cancel-btn"),
                    ),
                    Button(
                        "Delete",
                        type="button",
                        id="delete-chat-confirm-btn",
                        cls=(ButtonT.primary, "delete-chat-confirm-btn"),
                    ),
                    cls="delete-chat-modal-actions",
                ),
                cls="delete-chat-modal-card",
            ),
            id="delete-chat-modal",
            cls="delete-chat-modal",
            aria_hidden="true",
            role="dialog",
            aria_modal="true",
            tabindex="-1",
        )
    
    def _shell(request: Request):
        return Div(
            Div(id="sidebar-backdrop", cls="sidebar-backdrop"),
            _sidebar(request),
            _main_panel(),
            _delete_chat_modal(),
            cls="app-shell",
        )

    @app.route("/")
    def home(request: Request):
        if auth_required and not _is_authorized(request):
            request.session["next"] = "/"
            login_path = auth_paths["login_path"] if auth_paths else "/login"
            return RedirectResponse(login_path, status_code=303)
        return (
            Title(hero_title),
            Meta(name="viewport", content="width=device-width, initial-scale=1.0"),
            Body(
                _shell(request),
                cls="min-h-screen",
                data_import_prefix=IMPORT_PREFIX,
            ),
        )

    return app



# Agno integration for Pylogue
import copy
import inspect
import os
from typing import Any, Optional

from .common import (
    PYLOGUE_INSTRUCTIONS,
    compose_system_prompt as _compose_system_prompt_common,
    extract_user_from_context as _extract_user_from_context,
    format_tool_result_summary as _format_tool_result_summary,
    format_tool_status_done as _format_tool_status_done,
    format_tool_status_running as _format_tool_status_running,
    get_export_state as _get_export_state_common,
    load_prompt_state as _load_prompt_state_common,
    resolve_tool_html as _resolve_tool_html,
    sanitize_history_answer as _sanitize_history_answer,
    should_render_tool_result_raw as _should_render_tool_result_raw,
    wrap_tool_html as _wrap_tool_html,
)


def logfire_instrument_agno(
    *,
    write_token: str | None = None,
    endpoint: str = "https://logfire-us.pydantic.dev",
    service_name: str = "pylogue-agno",
    use_batch_processor: bool = True,
):
    """Instrument Agno with OpenTelemetry and export spans to Logfire via OTLP."""
    resolved_write_token = write_token or os.getenv("LOGFIRE_TOKEN") or os.getenv("LOGFIRE_WRITE_TOKEN")
    if not resolved_write_token:
        raise ValueError(
            "Missing Logfire write token. Set LOGFIRE_TOKEN (or LOGFIRE_WRITE_TOKEN) or pass write_token=..."
        )

    try:
        from openinference.instrumentation.agno import AgnoInstrumentor
    except ModuleNotFoundError as exc:
        raise RuntimeError(
            "Agno Logfire instrumentation requires `openinference-instrumentation-agno`."
        ) from exc

    try:
        from opentelemetry import trace as trace_api
        from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
        from opentelemetry.sdk.resources import Resource
        from opentelemetry.sdk.trace import TracerProvider
        from opentelemetry.sdk.trace.export import BatchSpanProcessor, SimpleSpanProcessor
    except ModuleNotFoundError as exc:
        raise RuntimeError(
            "Agno Logfire instrumentation requires OpenTelemetry SDK packages."
        ) from exc

    resolved_endpoint = endpoint.rstrip("/")
    resource = Resource.create({"service.name": service_name}) if service_name else None
    tracer_provider = TracerProvider(resource=resource)
    span_processor_cls = BatchSpanProcessor if use_batch_processor else SimpleSpanProcessor
    tracer_provider.add_span_processor(
        span_processor_cls(
            OTLPSpanExporter(
                endpoint=f"{resolved_endpoint}/v1/traces",
                headers={"Authorization": resolved_write_token},
            )
        )
    )
    trace_api.set_tracer_provider(tracer_provider)
    AgnoInstrumentor().instrument()
    return tracer_provider


def _normalize_tool_payload(tool_entry: Any) -> tuple[str | None, Any, Any, str | None]:
    if isinstance(tool_entry, dict):
        function = tool_entry.get("function")
        if isinstance(function, dict):
            function_name = function.get("name")
            function_args = function.get("arguments")
        else:
            function_name = None
            function_args = None

        tool_name = (
            tool_entry.get("tool_name")
            or tool_entry.get("name")
            or function_name
            or tool_entry.get("tool")
        )
        args = (
            tool_entry.get("tool_args")
            or tool_entry.get("args")
            or tool_entry.get("arguments")
            or function_args
            or tool_entry.get("input")
        )
        result = (
            tool_entry.get("result")
            or tool_entry.get("output")
            or tool_entry.get("content")
            or tool_entry.get("observation")
        )
        call_id = tool_entry.get("tool_call_id") or tool_entry.get("call_id") or tool_entry.get("id")
    else:
        tool_name = (
            getattr(tool_entry, "tool_name", None)
            or getattr(tool_entry, "name", None)
            or getattr(tool_entry, "tool", None)
        )
        args = (
            getattr(tool_entry, "tool_args", None)
            or getattr(tool_entry, "args", None)
            or getattr(tool_entry, "arguments", None)
            or getattr(tool_entry, "input", None)
        )
        result = (
            getattr(tool_entry, "result", None)
            or getattr(tool_entry, "output", None)
            or getattr(tool_entry, "content", None)
            or getattr(tool_entry, "observation", None)
        )
        call_id = (
            getattr(tool_entry, "tool_call_id", None)
            or getattr(tool_entry, "call_id", None)
            or getattr(tool_entry, "id", None)
        )
    return tool_name, args, result, call_id


def _extract_tools(chunk) -> list[tuple[str | None, Any, Any, str | None]]:
    payloads: list[Any] = []

    tools = getattr(chunk, "tools", None)
    if isinstance(tools, list):
        payloads.extend(tools)
    elif tools is not None:
        payloads.append(tools)

    # Agno ToolCallStarted/ToolCallCompleted events expose singular `tool`.
    tool = getattr(chunk, "tool", None)
    if tool is not None:
        payloads.append(tool)

    # Some providers can expose tool executions directly.
    tool_executions = getattr(chunk, "tool_executions", None)
    if isinstance(tool_executions, list):
        payloads.extend(tool_executions)
    elif tool_executions is not None:
        payloads.append(tool_executions)

    return [_normalize_tool_payload(t) for t in payloads]


def _is_event(event_value, expected_name: str) -> bool:
    if event_value is None:
        return False
    event_str = str(event_value)
    return event_str == expected_name or event_str.endswith(expected_name)


def _extract_content_text(chunk: Any) -> str | None:
    content = getattr(chunk, "content", None)
    if content is None:
        return None
    if isinstance(content, str):
        return content
    return str(content)


def _extract_reasoning_text(chunk: Any) -> str | None:
    reasoning_content = getattr(chunk, "reasoning_content", None)
    if reasoning_content is None:
        return None
    if isinstance(reasoning_content, str):
        return reasoning_content
    return str(reasoning_content)


def _message_to_dict(message: Any) -> dict | None:
    if isinstance(message, dict):
        role = message.get("role")
        content = message.get("content")
        if role is None or content is None:
            return None
        return {"role": role, "content": str(content)}
    role = getattr(message, "role", None)
    content = getattr(message, "content", None)
    if role is None or content is None:
        return None
    return {"role": str(role), "content": str(content)}


def _normalize_history_messages(messages: Any) -> list[dict]:
    if not isinstance(messages, list):
        return []
    normalized = []
    for message in messages:
        converted = _message_to_dict(message)
        if converted is not None:
            normalized.append(converted)
    return normalized


class AgnoResponder:
    """Streaming responder using Agno's stream events."""

    pylogue_instructions = PYLOGUE_INSTRUCTIONS

    def __init__(
        self,
        agent: Any,
        show_tool_details: bool = True,
        run_kwargs: Optional[dict[str, Any]] = None,
    ):
        self.agent = agent
        existing_instructions = getattr(agent, "instructions", None)
        if isinstance(existing_instructions, str):
            base_prompt = existing_instructions
        elif isinstance(existing_instructions, (list, tuple)):
            base_prompt = "\n".join(str(p) for p in existing_instructions if p is not None)
        else:
            base_prompt = ""

        state = getattr(agent, "_pylogue_prompt_state", None)
        if state is None:
            state = {
                "base_prompt": base_prompt,
                "additional": [],
            }
            agent._pylogue_prompt_state = state

        self._prompt_state = state
        self.message_history = []
        self.show_tool_details = show_tool_details
        self._active_user = None
        self.run_kwargs = copy.deepcopy(run_kwargs) if isinstance(run_kwargs, dict) else {}

    def append_instructions(self, additional_instructions: str) -> None:
        if additional_instructions:
            self._prompt_state["additional"].append(additional_instructions)

    def _compose_system_prompt(self, user: Optional[dict] = None) -> str:
        return _compose_system_prompt_common(
            base_prompt=self._prompt_state.get("base_prompt", ""),
            additional_instructions=list(self._prompt_state.get("additional", [])),
            user=user,
            pylogue_instructions=self.pylogue_instructions,
        )

    def get_export_state(self) -> dict:
        return _get_export_state_common(self._prompt_state, self._compose_system_prompt())

    def load_state(self, meta: dict) -> None:
        _load_prompt_state_common(self._prompt_state, meta)

    def load_history(self, cards, context=None) -> None:
        history = []
        for card in cards or []:
            question = card.get("question")
            answer = card.get("answer")
            if question is not None:
                history.append({"role": "user", "content": str(question)})
            answer_text = card.get("answer_text") if isinstance(card, dict) else None
            if answer_text is None:
                answer_text = answer
            answer_text = _sanitize_history_answer(answer_text)
            if answer_text:
                history.append({"role": "assistant", "content": str(answer_text)})
        self.message_history = history
        self.set_context(context)

    def set_context(self, context=None) -> None:
        self._active_user = _extract_user_from_context(context)

    def _build_run_input(self, text: str):
        history = list(self.message_history or [])
        history.append({"role": "user", "content": str(text)})
        return history

    def _build_run_kwargs(self) -> dict[str, Any]:
        kwargs = copy.deepcopy(self.run_kwargs)
        user = self._active_user
        if isinstance(user, dict) and "user_id" not in kwargs:
            user_id = user.get("email") or user.get("display_name") or user.get("name")
            if isinstance(user_id, str) and user_id:
                kwargs["user_id"] = user_id
        return kwargs

    async def __call__(self, text: str, context=None):
        self.set_context(context)
        user = self._active_user
        try:
            run_call_result = self.agent.arun(
                self._build_run_input(text),
                stream=True,
                stream_events=True,
                additional_context=self._compose_system_prompt(user=user),
                **self._build_run_kwargs(),
            )
            if inspect.isawaitable(run_call_result):
                run_stream = await run_call_result
            else:
                run_stream = run_call_result
        except ModuleNotFoundError as exc:
            raise RuntimeError("Agno integration requires the `agno` package. Install it with `pip install agno`.") from exc
        except AttributeError as exc:
            raise RuntimeError("Configured Agno agent does not expose `arun`.") from exc

        if not hasattr(run_stream, "__aiter__"):
            raise TypeError(
                "Agno `arun(..., stream=True, stream_events=True)` must return an async iterable."
            )

        pending_tool_calls: dict[str, tuple[str | None, Any]] = {}
        tool_call_counter = 0
        streamed_text = ""
        history_updated_from_stream = False

        async for chunk in run_stream:
            event = getattr(chunk, "event", None)

            if _is_event(event, "RunContent") or _is_event(event, "RunIntermediateContent") or _is_event(event, "RunResponse") or _is_event(event, "RunCompleted"):
                content = _extract_content_text(chunk)
                if content:
                    if content.startswith(streamed_text):
                        delta = content[len(streamed_text) :]
                        streamed_text = content
                    else:
                        delta = content
                        streamed_text += content
                    if delta:
                        yield delta
                continue

            if _is_event(event, "ToolCallStarted"):
                for tool_name, args, _, call_id in _extract_tools(chunk):
                    tool_call_counter += 1
                    resolved_call_id = call_id or f"tool-{tool_call_counter}"
                    pending_tool_calls[resolved_call_id] = (tool_name, args)
                    if not self.show_tool_details:
                        yield _format_tool_status_running(tool_name, args, resolved_call_id)
                continue

            if _is_event(event, "ToolCallCompleted"):
                for tool_name, args, result, call_id in _extract_tools(chunk):
                    resolved_call_id = call_id
                    if resolved_call_id in pending_tool_calls:
                        tool_name, args = pending_tool_calls.pop(resolved_call_id)
                    if self.show_tool_details and (tool_name or args or result):
                        yield _format_tool_result_summary(tool_name, args, result)
                    if not self.show_tool_details:
                        yield _format_tool_status_done(args, resolved_call_id, tool_name)
                    resolved_html = _resolve_tool_html(result)
                    if resolved_html:
                        yield _wrap_tool_html(resolved_html)
                    elif _should_render_tool_result_raw(tool_name, result):
                        yield _wrap_tool_html(result)
                continue

            reasoning_text = _extract_reasoning_text(chunk)
            if reasoning_text and (_is_event(event, "ReasoningContentDelta") or _is_event(event, "Reasoning")):
                yield reasoning_text

            content = _extract_content_text(chunk)
            if content and _is_event(event, "Reasoning"):
                yield content

            messages = _normalize_history_messages(getattr(chunk, "messages", None))
            if messages:
                self.message_history = messages
                history_updated_from_stream = True

        if pending_tool_calls and self.show_tool_details:
            for tool_name, args in pending_tool_calls.values():
                yield _format_tool_result_summary(tool_name, args, None)

        if not history_updated_from_stream:
            updated_history = list(self.message_history) if isinstance(self.message_history, list) else []
            updated_history.append({"role": "user", "content": str(text)})
            updated_history.append({"role": "assistant", "content": streamed_text})
            self.message_history = updated_history



import html
import json
import re
import ast
from typing import Any, Optional

PYLOGUE_INSTRUCTIONS = (
    "You are also a helpful AI assistant integrated with Pylogue enviroment."
    "The enviroment supports auto injection of html, i.e., if you respond with raw HTML it will be rendered as HTML."
    "The environment also supports markdown rendering, so you can use markdown syntax for formatting."
    "Finally the environment supports mermaid diagrams, so you can create diagrams using mermaid syntax. with ```mermaid ... ``` blocks."
    "Always generate (block appropriate) css based colorful mermaid diagrams (e.g., classDef evaporation fill:#add8e6,stroke:#333,stroke-width:2px) when appropriate to illustrate concepts."
    "also ensure in mermaid blocks you wrap the text with double quotes to avoid syntax errors, and <br> for line breaks instead of \\n"
    "prefer vertical layouts for flowcharts and sequence diagrams. "
    "Render math using LaTeX syntax within $$ ... $$ blocks or inline with $ ... $."
    "when embedding HTML do not wrap it inside ```html ... ``` blocks, just output the raw HTML directly. Do not add <html> or <body> tags."
    "Just because you can respond with HTML or generate mermaid diagrams does not mean you should always do that. Apart from accuracy of response, your next biggest goals is to save as many tokens as possible while ensuring the response is clear and complete."
)

_TOOL_HTML_RE = re.compile(r'<div class="tool-html">.*?</div>', re.DOTALL)
_TAG_RE = re.compile(r"<[^>]+>")


def sanitize_history_answer(answer: str) -> str:
    if not isinstance(answer, str) or not answer:
        return ""
    text = _TOOL_HTML_RE.sub("Rendered tool output.", answer)
    text = _TAG_RE.sub("", text)
    return html.unescape(text).strip()


def safe_json(value):
    if value is None:
        return "{}"
    if isinstance(value, str):
        try:
            return json.dumps(json.loads(value), indent=2, sort_keys=True, ensure_ascii=True)
        except json.JSONDecodeError:
            return value
    try:
        return json.dumps(value, indent=2, sort_keys=True, ensure_ascii=True)
    except TypeError:
        return json.dumps(str(value), indent=2, sort_keys=True, ensure_ascii=True)


def truncate(text: str, limit: int = 100) -> str:
    if not isinstance(text, str):
        return ""
    return text if len(text) <= limit else f"{text[:limit//2]} ... (truncated) ... {text[-limit//2:]}"


def safe_dom_id(value: str | None) -> str:
    if not value:
        return "tool-status"
    safe = []
    for ch in str(value):
        if ch.isalnum() or ch in {"-", "_"}:
            safe.append(ch)
    return "".join(safe) or "tool-status"


def format_tool_result_summary(tool_name: str, args, result):
    tool_label = html.escape(tool_name or "tool")
    safe_args = html.escape(safe_json(args))
    safe_result = html.escape(truncate(safe_json(result)))
    return (
        "\n\n"
        f'<details class="tool-call"><summary>Tool: {tool_label}</summary>'
        f"<div><strong>Args</strong></div>"
        f"<pre><code>{safe_args}</code></pre>"
        f"<div><strong>Result</strong></div>"
        f"<pre><code>{safe_result}</code></pre></details>\n\n"
    )


def format_tool_status_running(tool_name: str, args, call_id: str | None):
    purpose = None
    if isinstance(args, dict):
        purpose = args.get("purpose")
    label = purpose or (tool_name.replace("_", " ").title() if tool_name else "Working")
    status_id = safe_dom_id(f"tool-status-{call_id or ''}")
    safe_label = html.escape(str(label))
    return (
        f'<div id="{status_id}" class="tool-status tool-status--running">{safe_label}</div>\n'
    )


def format_tool_status_done(args, call_id: str | None, tool_name: str | None = None):
    purpose = None
    if isinstance(args, dict):
        purpose = args.get("purpose")
    
    # Use the same fallback logic as format_tool_status_running
    if purpose and isinstance(purpose, str) and purpose.strip():
        safe_label = purpose.strip()
    elif tool_name:
        safe_label = tool_name.replace("_", " ").title()
    else:
        safe_label = "Task"
    
    status_id = safe_dom_id(f"tool-status-{call_id or ''}")
    safe_label_escaped = html.escape(safe_label)
    # Show process name + highlighted completion status
    return (
        f'<div class="tool-status-update" data-target-id="{status_id}">'
        f'{safe_label_escaped} <span class="tool-status-check">✓Completed</span></div>\n'
    )


def resolve_tool_html(result):
    parsed_result = result
    if isinstance(parsed_result, str):
        stripped = parsed_result.strip()
        if stripped.startswith("{") and stripped.endswith("}"):
            try:
                parsed_result = json.loads(stripped)
            except Exception:
                try:
                    parsed_result = ast.literal_eval(stripped)
                except Exception:
                    parsed_result = result
    if isinstance(parsed_result, dict) and "_pylogue_html_id" in parsed_result:
        token = parsed_result.get("_pylogue_html_id")
        try:
            from pylogue.embeds import take_html
        except Exception:
            return None
        return take_html(token)
    return None


def should_render_tool_result_raw(tool_name: str | None, result) -> bool:
    if not isinstance(result, str):
        return False
    stripped = result.lstrip()
    if not stripped.startswith("<"):
        return False
    return True


def wrap_tool_html(result: str) -> str:
    stripped = result.strip()
    if stripped.startswith("<div") and stripped.endswith("</div>"):
        return result
    return f'<div class="tool-html">{result}</div>'


def extract_user_from_context(context):
    if not isinstance(context, dict):
        return None
    user = context.get("user")
    return user if isinstance(user, dict) else None


def compose_system_prompt(
    base_prompt: str,
    additional_instructions: list[str],
    user: Optional[dict] = None,
    pylogue_instructions: str = PYLOGUE_INSTRUCTIONS,
) -> str:
    segments = []
    if base_prompt:
        segments.append(base_prompt)
    segments.append(pylogue_instructions)
    if isinstance(user, dict):
        display_name = user.get("display_name") or user.get("name")
        email = user.get("email")
        user_parts = []
        if display_name:
            user_parts.append(f"name={display_name}")
        if email:
            user_parts.append(f"email={email}")
        if user_parts:
            segments.append(
                "Authenticated user profile (source of truth): "
                + ", ".join(user_parts)
                + ". Use this identity when the user asks who they are or asks for personalization."
            )
    if additional_instructions:
        segments.extend(additional_instructions)
    return "\n\n".join(segments)


def get_export_state(prompt_state: dict, system_prompt: str) -> dict:
    return {
        "prompt_state": {
            "base_prompt": prompt_state.get("base_prompt", ""),
            "additional": list(prompt_state.get("additional", [])),
        },
        "system_prompt": system_prompt,
    }


def load_prompt_state(prompt_state: dict, meta: dict) -> None:
    if not isinstance(meta, dict):
        return
    exported_state = meta.get("prompt_state") if isinstance(meta.get("prompt_state"), dict) else {}
    if "base_prompt" in exported_state:
        prompt_state["base_prompt"] = exported_state.get("base_prompt") or ""
    if "additional" in exported_state and isinstance(exported_state.get("additional"), list):
        prompt_state["additional"] = list(exported_state.get("additional", []))
    elif isinstance(meta.get("system_prompt"), str):
        prompt_state["additional"] = [meta["system_prompt"]]



import html as html_lib
import json

from loguru import logger
from pylogue.embeds import store_html

try:
    import pandas as pd
except ImportError:
    pd = None

try:
    import plotly.express as px
    import plotly.graph_objects as go
    from plotly.subplots import make_subplots
    from plotly.utils import PlotlyJSONEncoder
except ImportError:
    px = None
    go = None
    make_subplots = None
    PlotlyJSONEncoder = None


def _preview(code: str, limit: int = 240) -> str:
    if not code:
        return ""
    compact = " ".join(code.strip().split())
    return compact[:limit] + ("..." if len(compact) > limit else "")


def render_plotly_chart_py(sql_query_runner: callable, sql_query: str, plotly_python: str):
    """Render a Plotly chart using Python code that defines `fig`.

    Prioritizes responsive behavior for chat UIs:
    - fills 100% available width of the chat card
    - adapts to browser resize and mobile breakpoints

    The code runs with access to:
    df (pandas DataFrame), pd (pandas), px (plotly.express),
    go (plotly.graph_objects), make_subplots (plotly.subplots.make_subplots).
    A list-of-dicts alias `data` is also provided for compatibility with
    snippets that start with `pd.DataFrame(data)`.

    Plotly dropdown safety rules (prevents blank charts after selection):
    - For `updatemenus` with `method="update"`, always provide per-trace arrays:
      `x: [series.tolist()]`, `y: [series.tolist()]`, `text: [series.tolist()]`.
    - For `customdata`, pass a 2D payload wrapped per trace:
      `customdata: [df[[...]].to_numpy()]` (never a bare DataFrame).
    - Keep trace type stable across updates, e.g. `fig.update_traces(type="bar", ...)`.
    - Keep `hovertemplate` stable when using `customdata` in dropdown-driven updates.
    - Place dropdowns in a stable area (not over the title), for example:
      `updatemenus=[dict(buttons=buttons, x=0.01, y=1.02, xanchor="left", yanchor="bottom")]`
      and use `fig.update_layout(margin=dict(t=120))`.
      For placement below the chart, use `y=-0.12` and increase `margin.b`.
    - Pylogue auto-normalizes dropdown payloads for common update mistakes,
      so all agents get safer behavior by default.
    - Optional cross-trace click interaction contract (tool-level, agent-agnostic):
      put `layout.meta.pylogue_linked_interaction` with:
      `source_trace` (int), `target_trace` (int), `lookup` (dict),
      and optional `season_menu_index` (int), `default_season` (str),
      `target_title_annotation_index` (int).
      `lookup` keys can be `"<season>||<x_label>"` or `"<x_label>"`.
      Each payload can include `x`, `y`, `text`, `customdata`, `title`.
    """

    try:
        if (
            pd is None
            or go is None
            or px is None
            or make_subplots is None
            or PlotlyJSONEncoder is None
        ):
            return (
                'Missing dependencies. Install with: '
                'pip install "pylogue[dashboard]" plotly'
            )

        local_scope = {
            "pd": pd,
            "px": px,
            "go": go,
            "make_subplots": make_subplots,
        }
        if sql_query_runner is not None and sql_query is not None:
            df = pd.DataFrame(sql_query_runner(sql_query))
            local_scope["df"] = df
            local_scope["data"] = df.to_dict(orient="records")

        try:
            logger.info(
                f"Executed Plotly code: sql_attached={bool(sql_query_runner and sql_query)}, code_preview\n---\n{plotly_python}\n---\n"
            )
            exec(plotly_python, local_scope)
        except Exception as exc:  # noqa: BLE001
            logger.exception(
                "Plotly code execution failed: sql_attached={}, code_preview={!r}",
                bool(sql_query_runner and sql_query),
                _preview(plotly_python),
            )
            return f"Error executing Plotly code: {exc}"

        fig = local_scope.get("fig")
        if fig is None or not hasattr(fig, "to_plotly_json"):
            return "Error: Plotly code must define a `fig` variable."

        try:
            fig_json = fig.to_plotly_json()
        except Exception as exc:  # noqa: BLE001
            logger.exception(
                "Plotly serialization failed: fig_type={}, code_preview={!r}",
                fig.__class__.__name__ if fig is not None else "unknown",
                _preview(plotly_python),
            )
            return f"Error serializing Plotly figure: {exc}"

        layout = fig_json.get("layout") or {}
        user_height = layout.get("height")
        has_explicit_height = isinstance(user_height, (int, float))
        default_height = int(user_height) if has_explicit_height else 420

        if not has_explicit_height:
            # Let Plotly autosize width while we manage a mobile-friendly height.
            fig_json.setdefault("layout", {})
            fig_json["layout"]["autosize"] = True
            fig_json["layout"].pop("width", None)

        fig_payload = json.dumps(fig_json, cls=PlotlyJSONEncoder)
        script_html = (
            "<div id='plot-wrap' style='width:100%;max-width:100%;margin:0;'>"
            "<div id='plot-root' style='width:100%;'></div>"
            "</div>"
            "<script src='https://cdn.plot.ly/plotly-2.35.2.min.js'></script>"
            "<script>"
            f"const fig={fig_payload};"
            f"const explicitHeight={'true' if has_explicit_height else 'false'};"
            f"const defaultHeight={default_height};"
            "const gd=document.getElementById('plot-root');"
            "const TRACE_KEYS=['x','y','text','customdata'];"
            "function normalizePerTraceValue(value, traceCount){"
            "if(value===undefined||value===null) return value;"
            "if(!Array.isArray(value)) return [value];"
            "if(traceCount===1){"
            "if(value.length===0) return [[]];"
            "if(Array.isArray(value[0])) return value;"
            "return [value];"
            "}"
            "if(value.length===traceCount && value.every((item)=>Array.isArray(item)||item===null)) return value;"
            "if(!Array.isArray(value[0])) return [value];"
            "return value;"
            "}"
            "function stripFragileAnnotationKeys(obj){"
            "if(!obj||typeof obj!=='object') return;"
            "for(const key of Object.keys(obj)){"
            "if(key.startsWith('annotations[')&&key.endsWith('].text')) delete obj[key];"
            "}"
            "}"
            "function normalizeUpdateMenus(){"
            "const layout=fig.layout||{};"
            "const menus=layout.updatemenus||[];"
            "const traceCount=(fig.data||[]).length||1;"
            "for(const menu of menus){"
            "const buttons=(menu&&menu.buttons)||[];"
            "for(const btn of buttons){"
            "if(!btn||!Array.isArray(btn.args)||btn.args.length===0) continue;"
            "if(String(btn.method||'').toLowerCase()!=='update') continue;"
            "const dataArgs=btn.args[0];"
            "if(dataArgs&&typeof dataArgs==='object'){"
            "for(const key of TRACE_KEYS){"
            "if(key in dataArgs){"
            "dataArgs[key]=normalizePerTraceValue(dataArgs[key], traceCount);"
            "}"
            "}"
            "stripFragileAnnotationKeys(dataArgs);"
            "}"
            "const layoutArgs=btn.args.length>1?btn.args[1]:null;"
            "if(layoutArgs&&typeof layoutArgs==='object'){"
            "stripFragileAnnotationKeys(layoutArgs);"
            "}"
            "}"
            "}"
            "}"
            "function setupLinkedInteraction(){"
            "const meta=((fig.layout||{}).meta||{}).pylogue_linked_interaction;"
            "if(!meta||typeof meta!=='object') return;"
            "const sourceTrace=Number.isInteger(meta.source_trace)?meta.source_trace:0;"
            "const targetTrace=Number.isInteger(meta.target_trace)?meta.target_trace:1;"
            "const seasonMenuIndex=Number.isInteger(meta.season_menu_index)?meta.season_menu_index:0;"
            "const lookup=(meta.lookup&&typeof meta.lookup==='object')?meta.lookup:{};"
            "const defaultSeason=meta.default_season==null?'':String(meta.default_season);"
            "function normalizeLinkedSeries(value){"
            "if(Array.isArray(value) && value.length===1 && Array.isArray(value[0])){"
            "return value[0];"
            "}"
            "return value;"
            "}"
            "function activeSeason(){"
            "try{"
            "const menus=(gd.layout&&gd.layout.updatemenus)||[];"
            "const menu=menus[seasonMenuIndex];"
            "if(!menu||!Array.isArray(menu.buttons)) return defaultSeason;"
            "const active=Number.isInteger(menu.active)?menu.active:0;"
            "const btn=menu.buttons[active];"
            "if(!btn||btn.label==null) return defaultSeason;"
            "return String(btn.label);"
            "}catch(_err){return defaultSeason;}"
            "}"
            "gd.on('plotly_click', function(eventData){"
            "const point=eventData&&eventData.points&&eventData.points[0];"
            "if(!point||point.curveNumber!==sourceTrace) return;"
            "const label=String(point.x);"
            "const season=activeSeason();"
            "const combinedKey=season+'||'+label;"
            "const payload=lookup[combinedKey]||lookup[label];"
            "if(!payload||typeof payload!=='object') return;"
            "const update={};"
            "for(const key of TRACE_KEYS){"
            "if(payload[key]!==undefined){"
            "const val=(key==='customdata')?payload[key]:normalizeLinkedSeries(payload[key]);"
            "update[key]=[val];"
            "}"
            "}"
            "if(Object.keys(update).length===0) return;"
            "try{"
            "const fullData=gd._fullData||[];"
            "const target=fullData[targetTrace]||{};"
            "const yRef=String(target.yaxis||'y');"
            "const yLayoutKey=(yRef==='y')?'yaxis':'yaxis'+yRef.slice(1);"
            "const yVals=payload.y;"
            "if(Array.isArray(yVals) && yVals.length>0 && typeof yVals[0]==='string'){"
            "const patch={};"
            "patch[yLayoutKey+'.type']='category';"
            "Plotly.relayout(gd, patch);"
            "}"
            "}catch(_err){}"
            "Plotly.restyle(gd, update, [targetTrace]);"
            "if(payload.title){"
            "try{"
            "const anns=(gd.layout&&gd.layout.annotations)||[];"
            "const idx=Number.isInteger(meta.target_title_annotation_index)?meta.target_title_annotation_index:(anns.length>1?1:0);"
            "if(anns.length && idx>=0 && idx<anns.length){"
            "const relayoutPatch={};"
            "relayoutPatch['annotations['+idx+'].text']=String(payload.title);"
            "Plotly.relayout(gd, relayoutPatch);"
            "}else{"
            "Plotly.relayout(gd, {title:String(payload.title)});"
            "}"
            "}catch(_err){}"
            "}"
            "});"
            "}"
            "function mobileHeight(){"
            "if(explicitHeight) return defaultHeight;"
            "const w=Math.max(280, Math.min(window.innerWidth||1024, 1400));"
            "return Math.max(280, Math.min(560, Math.round(w*0.6)));"
            "}"
            "function render(){"
            "normalizeUpdateMenus();"
            "fig.layout=fig.layout||{};"
            "fig.layout.autosize=true;"
            "fig.layout.width=null;"
            "fig.layout.height=mobileHeight();"
            "if(window.frameElement){"
            "window.frameElement.style.height=fig.layout.height+'px';"
            "window.frameElement.height=String(fig.layout.height);"
            "}"
            "Plotly.react(gd, fig.data||[], fig.layout, {responsive:true, displaylogo:false});"
            "}"
            "render();"
            "setupLinkedInteraction();"
            "const ro=new ResizeObserver(()=>{"
            "render();"
            "try{Plotly.Plots.resize(gd);}catch(e){}"
            "});"
            "ro.observe(document.body);"
            "window.addEventListener('resize', ()=>{"
            "render();"
            "try{Plotly.Plots.resize(gd);}catch(e){}"
            "});"
            "</script>"
        )

        srcdoc = (
            "<!doctype html><html><head><meta charset='utf-8'/>"
            "<meta name='viewport' content='width=device-width,initial-scale=1'/>"
            "<style>html,body{margin:0;padding:0;background:#fff;overflow:hidden;}"
            "#plot-wrap{width:100%;max-width:100%;}</style>"
            "</head><body>"
            f"{script_html}"
            "</body></html>"
        )
        escaped_srcdoc = html_lib.escape(srcdoc, quote=True)
        iframe_html = (
            '<iframe '
            f'srcdoc="{escaped_srcdoc}" '
            f'height="{default_height}" '
            "style=\"width:100%;max-width:100%;height:"
            f"{default_height}px;border:0;display:block;\" "
            'title="Plotly Chart"></iframe>'
        )
        html_id = store_html(iframe_html)
        return {"_pylogue_html_id": html_id, "message": "Plotly chart rendered."}
    except Exception as e:
        logger.exception(
            "Unhandled error in render_plotly_chart_py: sql_attached={}, code_preview={!r}",
            bool(sql_query_runner and sql_query),
            _preview(plotly_python),
        )
        return f"Error in render_plotly_chart_py: {e}"



# Core FastHTML + MonsterUI chat
from fasthtml.common import *
from monsterui.all import Theme, Container, ContainerT, TextPresets, Button, ButtonT, FastHTML as MUFastHTML, UkIcon
from dataclasses import dataclass
from pathlib import Path
from urllib.parse import quote_plus
from starlette.requests import Request
from starlette.responses import FileResponse, RedirectResponse
import asyncio
import inspect
import json
import base64
import html as html_lib
import logging
import os
import re

IMPORT_PREFIX = "__PYLOGUE_IMPORT__:"
STOP_PREFIX = "__PYLOGUE_STOP__:"
_CORE_STATIC_DIR = Path(__file__).resolve().parent / "static"
_LOG = logging.getLogger(__name__)


@dataclass(frozen=True)
class GoogleOAuthConfig:
    client_id: str
    client_secret: str
    allowed_domains: tuple[str, ...] = ()
    allowed_emails: tuple[str, ...] = ()
    auth_required: bool = True
    session_secret: str | None = None


def _split_csv_env(value: str | None) -> tuple[str, ...]:
    if not value:
        return ()
    return tuple(part.strip() for part in value.split(",") if part.strip())


def _env_bool(name: str, default: bool = False) -> bool:
    raw = os.getenv(name)
    if raw is None:
        return default
    return raw.strip().lower() in {"1", "true", "yes", "on"}


def google_oauth_config_from_env() -> GoogleOAuthConfig | None:
    client_id = os.getenv("PYLOGUE_GOOGLE_CLIENT_ID") or os.getenv("PYLOGUE_CLIENT_ID")
    client_secret = os.getenv("PYLOGUE_GOOGLE_CLIENT_SECRET") or os.getenv("PYLOGUE_CLIENT_SECRET")
    if not (client_id and client_secret):
        return None
    return GoogleOAuthConfig(
        client_id=client_id,
        client_secret=client_secret,
        allowed_domains=_split_csv_env(os.getenv("PYLOGUE_GOOGLE_ALLOWED_DOMAINS")),
        allowed_emails=_split_csv_env(os.getenv("PYLOGUE_GOOGLE_ALLOWED_EMAILS")),
        auth_required=_env_bool("PYLOGUE_AUTH_REQUIRED", default=True),
        session_secret=os.getenv("PYLOGUE_SESSION_SECRET"),
    )


def _normalize_base_path(base_path: str) -> str:
    base_path = (base_path or "").strip()
    if base_path in {"", "/"}:
        return ""
    normalized = "/" + base_path.strip("/")
    if ".." in normalized.split("/"):
        raise ValueError("base_path cannot contain '..'")
    return normalized


def _request_auth(request: Request):
    try:
        auth = request.session.get("auth")
    except Exception:
        auth = None
    if isinstance(auth, dict):
        return auth
    return None


def _connection_auth(conn):
    try:
        scope = conn.scope
    except Exception:
        scope = None
    if not isinstance(scope, dict):
        return None
    session = scope.get("session")
    if session is None:
        return None
    try:
        auth = session.get("auth")
    except Exception:
        auth = None
    if isinstance(auth, dict):
        return auth
    return None


def _user_context_from_auth(auth):
    if not isinstance(auth, dict):
        return None
    name = auth.get("name") or auth.get("username")
    email = auth.get("email")
    return {
        "name": name,
        "email": email,
        "display_name": name or email,
        "provider": auth.get("provider"),
    }


def _build_responder_context(conn):
    auth = _connection_auth(conn)
    if not auth:
        return None
    return {
        "auth": auth,
        "user": _user_context_from_auth(auth),
    }


def _invoke_responder(responder, prompt: str, context):
    try:
        signature = inspect.signature(responder)
    except (TypeError, ValueError):
        signature = None

    if signature is not None:
        params = signature.parameters
        if "context" in params or any(
            p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values()
        ):
            return responder(prompt, context=context)
    try:
        return responder(prompt)
    except TypeError:
        return responder(prompt, context)


def _oauth_base_url(request: Request) -> str:
    explicit = os.getenv("PYLOGUE_PUBLIC_URL")
    if explicit:
        return explicit.rstrip("/")
    base = str(request.base_url).rstrip("/")
    return base.replace("://0.0.0.0", "://localhost")


def _session_cookie_name() -> str:
    return os.getenv("PYLOGUE_SESSION_COOKIE", "pylogue_session")


def _register_google_auth_routes(app, cfg: GoogleOAuthConfig, base_path: str = "") -> dict[str, str]:
    try:
        from authlib.integrations.starlette_client import OAuth
    except Exception as exc:
        raise RuntimeError("Google OAuth requires authlib. Install with `pip install authlib`.") from exc

    oauth = OAuth()
    oauth.register(
        name="google",
        client_id=cfg.client_id,
        client_secret=cfg.client_secret,
        server_metadata_url="https://accounts.google.com/.well-known/openid-configuration",
        userinfo_endpoint="https://openidconnect.googleapis.com/v1/userinfo",
        client_kwargs={"scope": "openid email profile"},
    )

    base = _normalize_base_path(base_path)
    login_path = f"{base}/login" if base else "/login"
    login_google_path = f"{base}/login/google" if base else "/login/google"
    callback_path = f"{base}/auth/google/callback" if base else "/auth/google/callback"
    logout_path = f"{base}/logout" if base else "/logout"
    default_next = f"{base}/" if base else "/"

    @app.route(login_path, methods=["GET"])
    async def pylogue_google_login(request: Request):
        error = request.query_params.get("error")
        return Div(
            H2("Login", cls="uk-h2"),
            A(
                Span("Continue with Google", cls="text-sm font-semibold"),
                href=login_google_path,
                cls=(
                    "inline-flex items-center justify-center px-4 py-2 my-6 rounded-md "
                    "border border-slate-700 bg-slate-800 text-slate-100 hover:bg-slate-900 "
                    "hover:border-slate-900 transition-colors max-w-sm mx-auto"
                ),
            ),
            P(error, cls="text-red-500 mt-4") if error else None,
            cls="prose mx-auto mt-24 text-center",
        )

    @app.route(login_google_path)
    async def pylogue_google_login_redirect(request: Request):
        next_url = request.session.get("next") or request.query_params.get("next") or default_next
        request.session["next"] = next_url
        redirect_uri = _oauth_base_url(request) + callback_path
        return await oauth.google.authorize_redirect(request, redirect_uri)

    @app.route(callback_path)
    async def pylogue_google_callback(request: Request):
        try:
            token = await oauth.google.authorize_access_token(request)
            userinfo = token.get("userinfo")
            if not userinfo:
                try:
                    userinfo = await oauth.google.parse_id_token(request, token)
                except Exception:
                    # Fallback to userinfo endpoint when id_token is unavailable.
                    resp = await oauth.google.get("https://openidconnect.googleapis.com/v1/userinfo", token=token)
                    userinfo = resp.json()
        except Exception as exc:
            _LOG.exception("Google OAuth callback failed: %s", exc)
            err = quote_plus(f"Google authentication failed ({type(exc).__name__})")
            return RedirectResponse(f"{login_path}?error={err}", status_code=303)

        email = userinfo.get("email") if isinstance(userinfo, dict) else None
        if not email:
            return RedirectResponse(f"{login_path}?error=Google+authentication+failed+(no+email)", status_code=303)
        if cfg.allowed_domains:
            if not email:
                return RedirectResponse(f"{login_path}?error=Google+account+not+allowed", status_code=303)
            domain = email.split("@")[-1]
            if domain not in cfg.allowed_domains:
                return RedirectResponse(f"{login_path}?error=Google+account+not+allowed", status_code=303)
        if cfg.allowed_emails and (not email or email not in cfg.allowed_emails):
            return RedirectResponse(f"{login_path}?error=Google+account+not+allowed", status_code=303)

        request.session["auth"] = {
            "provider": "google",
            "email": email,
            "name": userinfo.get("name") if isinstance(userinfo, dict) else None,
            "picture": userinfo.get("picture") if isinstance(userinfo, dict) else None,
        }
        next_url = request.session.pop("next", default_next)
        return RedirectResponse(next_url, status_code=303)

    @app.route(logout_path)
    async def pylogue_google_logout(request: Request):
        request.session.pop("auth", None)
        request.session.pop("next", None)
        return RedirectResponse(login_path, status_code=303)

    return {
        "login_path": login_path,
        "logout_path": logout_path,
        "default_next": default_next,
    }


def register_core_static(app):
    if getattr(app, "_pylogue_static_registered", False):
        return
    app._pylogue_static_registered = True

    @app.route("/static/pylogue-core.css")
    def _pylogue_core_css():
        return FileResponse(_CORE_STATIC_DIR / "pylogue-core.css")

    @app.route("/static/pylogue-core.js")
    def _pylogue_core_js():
        return FileResponse(_CORE_STATIC_DIR / "pylogue-core.js")

    @app.route("/static/pylogue-markdown.js")
    def _pylogue_markdown_js():
        return FileResponse(_CORE_STATIC_DIR / "pylogue-markdown.js")

class EchoResponder:
    async def __call__(self, message: str, context=None):
        user = context.get("user") if isinstance(context, dict) else None
        display_name = user.get("display_name") if isinstance(user, dict) else None
        if display_name:
            response = f"[ECHO] {display_name}:\n{message}"
        else:
            response = f"[ECHO]:\n{message}"
        for ch in response:
            await asyncio.sleep(0.005)
            yield ch


def render_input():
    return Textarea(
        id="msg",
        name="msg",
        placeholder="Say hi...",
        autofocus=True,
        rows=1,
        cls="uk-textarea w-full bg-white border-slate-300 focus:border-slate-500 focus:ring-2 focus:ring-slate-200 font-mono",
    )


def render_cards(cards):
    rows = []
    data_json = json.dumps(cards)
    for card in cards:
        card_id = card.get("id", "")
        assistant_id = f"assistant-{card_id}" if card_id else ""
        rows.append(
            Div(
                Div(
                    NotStr('<svg width="27" height="27" viewBox="0 0 27 27" fill="none" xmlns="http://www.w3.org/2000/svg" class="user-icon"><circle cx="6.54545" cy="6.54545" r="6.54545" fill="#8FC187"/><circle cx="6.54545" cy="19.6364" r="6.54545" fill="#8FC187"/><circle cx="19.6364" cy="6.54545" r="6.54545" fill="#8FC187"/><circle cx="19.6364" cy="19.6364" r="6.54545" fill="#8FC187"/></svg>'),
                    Div(
                        card["question"],
                        data_raw_b64=base64.b64encode(card["question"].encode("utf-8")).decode("ascii"),
                        cls="marked text-base",
                    ),
                    cls="user-message-content",
                ),
                cls="chat-row-block chat-row-user",
            )
        )
        rows.append(
            Div(
                Div(
                    Button(
                        UkIcon("copy"),
                        cls="uk-button uk-button-text copy-btn",
                        type="button",
                        data_copy_target=assistant_id,
                        aria_label="Copy response",
                        title="Copy response",
                    ),
                    cls="flex justify-end",
                ),
                Div(
                    card["answer"] or "…",
                    id=assistant_id if assistant_id else None,
                    data_raw_b64=base64.b64encode((card["answer"] or "").encode("utf-8")).decode("ascii"),
                    cls="marked text-base text-left",
                ),
                cls="chat-row-block chat-row-assistant",
            )
        )
    
    # Only wrap in chat-panel if there are cards to display
    panel_content = [
        *rows,
        Div(id="scroll-anchor"),
        Input(type="hidden", id="chat-data", value=data_json),
        Input(type="hidden", id="chat-export", value=json.dumps({"cards": cards})),
    ]
    
    if cards:
        return Div(
            Div(
                *panel_content,
            ),
            id="cards",
            cls="chat-panel",
        )
    else:
        return Div(
            *panel_content,
            id="cards",
        )


def render_chat_data(cards):
    return Input(
        type="hidden",
        id="chat-data",
        value=json.dumps(cards),
        hx_swap_oob="true",
    )

_TOOL_HTML_RE = re.compile(r'<div class="tool-html">.*?</div>', re.DOTALL)
_TAG_RE = re.compile(r"<[^>]+>")


def _normalize_answer_for_history(answer: str) -> str:
    if not isinstance(answer, str) or not answer:
        return ""
    text = _TOOL_HTML_RE.sub("Rendered tool output.", answer)
    text = _TAG_RE.sub("", text)
    return html_lib.unescape(text).strip()


def build_export_payload(cards, responder=None):
    export_cards = []
    for card in cards:
        if not isinstance(card, dict):
            continue
        answer = card.get("answer", "")
        answer_text = card.get("answer_text")
        if not isinstance(answer_text, str) or not answer_text.strip():
            answer_text = _normalize_answer_for_history(answer)
        export_card = dict(card)
        export_card["answer_text"] = answer_text
        export_cards.append(export_card)

    payload = {"cards": export_cards}
    if responder is not None and hasattr(responder, "get_export_state"):
        try:
            meta = responder.get_export_state()
        except Exception:
            meta = None
        if meta:
            payload["meta"] = meta
    return payload


def render_chat_export(cards, responder=None):
    payload = build_export_payload(cards, responder=responder)
    return Input(
        type="hidden",
        id="chat-export",
        value=json.dumps(payload),
        hx_swap_oob="true",
    )



def render_assistant_update(card):
    card_id = card.get("id", "")
    assistant_id = f"assistant-{card_id}" if card_id else ""
    return Div(
        card.get("answer", "") or "…",
        id=assistant_id if assistant_id else None,
        data_raw_b64=base64.b64encode((card.get("answer", "") or "").encode("utf-8")).decode("ascii"),
        cls="marked text-base text-left",
        hx_swap_oob="true",
    )


def get_core_headers(include_markdown: bool = True):
    headers = list(Theme.slate.headers())
    if include_markdown:
        headers.extend(
            [
                Link(rel="stylesheet", href="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/katex.min.css"),
                Link(
                    rel="stylesheet",
                    href="https://cdn.jsdelivr.net/npm/@highlightjs/cdn-assets@11.9.0/styles/github.min.css",
                ),
                Script(src="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/katex.min.js"),
                Script(src="https://cdn.jsdelivr.net/npm/katex@0.16.9/dist/contrib/auto-render.min.js"),
                Script(src="https://cdn.jsdelivr.net/npm/vega@5"),
                Script(src="https://cdn.jsdelivr.net/npm/vega-lite@5"),
                Script(src="https://cdn.jsdelivr.net/npm/vega-embed@6"),
            ]
        )
        headers.append(
            Script(src="https://cdn.jsdelivr.net/npm/@highlightjs/cdn-assets@11.9.0/highlight.min.js")
        )
        headers.append(Script(src="/static/pylogue-markdown.js", type="module"))

    headers.append(Link(rel="stylesheet", href="/static/pylogue-core.css"))
    headers.append(Script(src="/static/pylogue-core.js", type="module"))

    return headers


def register_ws_routes(
    app,
    responder=None,
    responder_factory=None,
    base_path: str = "",
    sessions: dict | None = None,
    auth_required: bool = False,
):
    if responder_factory is None:
        responder = responder or EchoResponder()
    base_path = _normalize_base_path(base_path)
    ws_path = f"{base_path}/ws" if base_path else "/ws"
    if sessions is None:
        sessions = {}

    def _on_connect(ws, send):
        if auth_required and not _connection_auth(ws):
            return
        session_context = _build_responder_context(ws)
        session_responder = responder_factory() if responder_factory else responder
        if hasattr(session_responder, "set_context"):
            try:
                session_responder.set_context(session_context)
            except Exception:
                pass
        ws_id = id(ws)
        sessions[ws_id] = {
            "cards": [],
            "responder": session_responder,
            "task": None,
            "context": session_context,
        }

    def _on_disconnect(ws):
        session = sessions.pop(id(ws), None)
        if session is None:
            return
        task = session.get("task")
        if task is not None and not task.done():
            task.cancel()

    @app.ws(ws_path, conn=_on_connect, disconn=_on_disconnect)
    async def ws_handler(msg: str, send, ws):
        if auth_required and not _connection_auth(ws):
            return
        ws_id = id(ws)
        session = sessions.get(ws_id)
        if session is None:
            session_context = _build_responder_context(ws)
            session_responder = responder_factory() if responder_factory else responder
            if hasattr(session_responder, "set_context"):
                try:
                    session_responder.set_context(session_context)
                except Exception:
                    pass
            session = {
                "cards": [],
                "responder": session_responder,
                "task": None,
                "context": session_context,
            }
            sessions[ws_id] = session
        cards = session["cards"]
        session_responder = session["responder"]
        current_task = session.get("task")
        context = _build_responder_context(ws)
        if context is not None:
            session["context"] = context
            if hasattr(session_responder, "set_context"):
                try:
                    session_responder.set_context(context)
                except Exception:
                    pass

        async def _run_message(prompt: str):
            cards.append({"id": str(len(cards)), "question": prompt, "answer": ""})
            await send(render_cards(cards))
            try:
                result = _invoke_responder(
                    session_responder,
                    prompt,
                    context=session.get("context"),
                )
                if inspect.isasyncgen(result):
                    async for chunk in result:
                        cards[-1]["answer"] += str(chunk)
                        await send(render_assistant_update(cards[-1]))
                else:
                    if inspect.isawaitable(result):
                        result = await result
                    for ch in str(result):
                        cards[-1]["answer"] += ch
                        await send(render_assistant_update(cards[-1]))
            except asyncio.CancelledError:
                if cards and cards[-1].get("answer"):
                    cards[-1]["answer"] += "\n\n[Stopped]"
                else:
                    cards[-1]["answer"] = "[Stopped]"
                await send(render_assistant_update(cards[-1]))
            finally:
                await send(render_chat_data(cards))
                await send(render_chat_export(cards, responder=session_responder))
                session["task"] = None
            return

        if isinstance(msg, str) and msg.startswith(IMPORT_PREFIX):
            if current_task is not None and not current_task.done():
                current_task.cancel()
            payload = msg[len(IMPORT_PREFIX) :].strip()
            try:
                imported = json.loads(payload) if payload else []
            except json.JSONDecodeError:
                imported = []
            meta = None
            if isinstance(imported, dict):
                meta = imported.get("meta")
                imported = imported.get("cards", [])
            normalized = []
            if isinstance(imported, list):
                if imported and all(isinstance(item, dict) and "role" in item for item in imported):
                    pending_question = None
                    for item in imported:
                        role = item.get("role")
                        content = item.get("content", "")
                        if role == "User":
                            pending_question = content
                        elif role == "Assistant":
                            if pending_question is None:
                                continue
                            normalized.append(
                                {
                                    "id": str(len(normalized)),
                                    "question": pending_question,
                                    "answer": content,
                                }
                            )
                            pending_question = None
                else:
                    for item in imported:
                        if not isinstance(item, dict):
                            continue
                        question = item.get("question")
                        answer = item.get("answer")
                        answer_text = item.get("answer_text")
                        if question is None or answer is None:
                            continue
                        normalized.append(
                            {
                                "id": str(len(normalized)),
                                "question": str(question),
                                "answer": str(answer),
                                "answer_text": str(answer_text) if answer_text is not None else None,
                            }
                        )
            session["cards"] = normalized
            if meta is not None and hasattr(session_responder, "load_state"):
                try:
                    session_responder.load_state(meta)
                except Exception:
                    pass
            if hasattr(session_responder, "load_history"):
                try:
                    session_responder.load_history(normalized, context=session.get("context"))
                except Exception:
                    pass
            await send(render_cards(normalized))
            await send(render_chat_data(normalized))
            await send(render_chat_export(normalized, responder=session_responder))
            return

        if isinstance(msg, str) and msg.startswith(STOP_PREFIX):
            if current_task is not None and not current_task.done():
                current_task.cancel()
            return

        if current_task is not None and not current_task.done():
            current_task.cancel()

        session["task"] = asyncio.create_task(_run_message(msg))
        return

    return sessions


def register_routes(
    app,
    responder=None,
    responder_factory=None,
    tag_line: str = "STREAMING DEMO",
    title: str = "Minimal Stream Chat",
    subtitle: str = "One question, one answer card. Response streams character-by-character.",
    base_path: str = "",
    inject_headers: bool = False,
    include_markdown: bool = True,
    tag_line_href: str = "",
    google_oauth_config: GoogleOAuthConfig | None = None,
    auth_required: bool | None = None,
):
    if responder_factory is None and responder is not None and hasattr(responder, "message_history"):
        raise ValueError(
            "Responder appears to be stateful (has message_history). "
            "Pass responder_factory to create a fresh responder per connection."
        )
    register_core_static(app)
    base_path = _normalize_base_path(base_path)
    chat_path = f"{base_path}/" if base_path else "/"
    ws_path = f"{base_path}/ws" if base_path else "/ws"

    oauth_cfg = google_oauth_config or google_oauth_config_from_env()
    if auth_required is None:
        auth_required = bool(oauth_cfg and oauth_cfg.auth_required)
    auth_paths = None
    if oauth_cfg:
        auth_paths = _register_google_auth_routes(app, oauth_cfg, base_path=base_path)
    elif auth_required:
        raise ValueError("auth_required=True needs google_oauth_config or PYLOGUE_GOOGLE_* env vars.")

    if inject_headers:
        for header in get_core_headers(include_markdown=include_markdown):
            app.hdrs = (*app.hdrs, header)

    if responder_factory is None:
        responder = responder or EchoResponder()
    register_ws_routes(
        app,
        responder=responder,
        responder_factory=responder_factory,
        base_path=base_path,
        auth_required=auth_required,
    )

    @app.route(chat_path)
    def home(request: Request):
        auth = _request_auth(request)
        if auth_required and not auth:
            request.session["next"] = chat_path
            login_path = auth_paths["login_path"] if auth_paths else "/login"
            return RedirectResponse(f"{login_path}?next={quote_plus(chat_path)}", status_code=303)

        tag_line_node = (
            A(
                tag_line,
                href=tag_line_href,
                cls="text-xs uppercase tracking-widest text-slate-500 hover:text-slate-700",
            )
            if tag_line_href
            else P(tag_line, cls="text-xs uppercase tracking-widest text-slate-500")
        )
        user_email = auth.get("email") if isinstance(auth, dict) else None
        logout_href = auth_paths["logout_path"] if auth_paths else "/logout"
        auth_bar = (
            Div(
                P(user_email or "Signed in", cls="text-xs text-slate-500"),
                A("Sign out", href=logout_href, cls="text-xs text-slate-700 hover:text-slate-900 underline"),
                cls="flex items-center justify-end gap-3",
            )
            if auth
            else None
        )
        return (
            Title(title),
            Meta(name="viewport", content="width=device-width, initial-scale=1.0"),
            Body(
                Container(
                    Div(
                        auth_bar,
                        Div(
                            tag_line_node,
                            H1(title, cls="text-3xl md:text-4xl font-semibold text-slate-900"),
                            P(subtitle, cls=(TextPresets.muted_sm, "text-slate-600")),
                            cls="space-y-2",
                        ),
                        Div(
                            Button(
                                UkIcon("download"),
                                cls="uk-button uk-button-text copy-chat-btn",
                                type="button",
                                aria_label="Download conversation JSON",
                                title="Download conversation JSON",
                            ),
                            Button(
                                UkIcon("upload"),
                                cls="uk-button uk-button-text copy-chat-btn upload-chat-btn",
                                type="button",
                                aria_label="Upload conversation JSON",
                                title="Upload conversation JSON",
                            ),
                            Input(
                                type="file",
                                id="chat-upload",
                                accept="application/json",
                                cls="sr-only",
                            ),
                            cls="flex justify-end gap-2",
                        ),
                        Div(
                            Div(render_cards([])),
                            Form(
                                render_input(),
                                Div(
                                    Button("Send", cls=ButtonT.primary, type="submit", id="chat-send-btn"),
                                    cls="flex items-center",
                                ),
                                id="form",
                                hx_ext="ws",
                                ws_connect=ws_path,
                                ws_send=True,
                                hx_target="#cards",
                                hx_swap="outerHTML",
                                cls="flex flex-col sm:flex-row gap-3 items-stretch pt-4",
                            ),
                            cls="chat-panel space-y-4",
                        ),
                        cls="space-y-6",
                    ),
                    cls=(ContainerT.lg, "py-10"),
                ),
                cls="min-h-screen bg-slate-50 text-slate-900",
            ),
        )

def main(
    responder=None,
    responder_factory=None,
    tag_line: str = "STREAMING DEMO",
    title: str = "Minimal Stream Chat",
    subtitle: str = "One question, one answer card. Response streams character-by-character.",
    include_markdown: bool = True,
    tag_line_href: str = "",
    google_oauth_config: GoogleOAuthConfig | None = None,
    auth_required: bool | None = None,
):
    if responder is None:
        responder = EchoResponder()
    headers = get_core_headers(include_markdown=include_markdown)
    oauth_cfg = google_oauth_config or google_oauth_config_from_env()
    session_secret = (
        oauth_cfg.session_secret
        if oauth_cfg and oauth_cfg.session_secret
        else os.getenv("PYLOGUE_SESSION_SECRET")
    )
    app_kwargs = {"exts": "ws", "hdrs": tuple(headers), "pico": False}
    app_kwargs["session_cookie"] = _session_cookie_name()
    if session_secret:
        app_kwargs["secret_key"] = session_secret
    app = MUFastHTML(**app_kwargs)
    register_routes(
        app,
        responder=responder,
        responder_factory=responder_factory,
        tag_line=tag_line,
        title=title,
        subtitle=subtitle,
        tag_line_href=tag_line_href,
        base_path="",
        google_oauth_config=oauth_cfg,
        auth_required=auth_required,
    )
    return app


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("pylogue.core:main", host="0.0.0.0", port=5001, reload=True, factory=True)



import secrets
import time
from typing import Optional

_HTML_CACHE: dict[str, tuple[float, str]] = {}
_TTL_SECONDS = 60 * 10


def _purge_expired(now: float) -> None:
    expired = [key for key, (ts, _) in _HTML_CACHE.items() if now - ts > _TTL_SECONDS]
    for key in expired:
        _HTML_CACHE.pop(key, None)


def store_html(html: str) -> str:
    """Store HTML and return a short-lived token."""
    now = time.time()
    _purge_expired(now)
    token = secrets.token_urlsafe(16)
    _HTML_CACHE[token] = (now, html)
    return token


def take_html(token: str) -> Optional[str]:
    """Retrieve and remove HTML by token."""
    if not token:
        return None
    entry = _HTML_CACHE.pop(token, None)
    if not entry:
        return None
    _, html = entry
    return html
